import os, json, shutil, traceback 
from collections import OrderedDict
from jinja2 import Environment, FileSystemLoader, PackageLoader, meta

from prompt_toolkit import prompt
from prompt_toolkit.token import Token
from prompt_toolkit.contrib.completers import PathCompleter
from prompt_toolkit.styles import style_from_dict
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.contrib.completers import WordCompleter

##########################
# Jinja2 helper functions...
##########################
def render(tpl_path, context):
    path, filename = os.path.split(tpl_path)
    return Environment(
        loader=FileSystemLoader(path or './')
    ).get_template(filename).render(context)

def get_variables(tpl_path):     
    path, filename = os.path.split(tpl_path)
    env = Environment(loader=FileSystemLoader(path or './'))
    template_source = env.loader.get_source(env, filename)[0]
    parsed_content = env.parse(template_source)
    return meta.find_undeclared_variables(parsed_content)

style = style_from_dict({
    Token.Toolbar: '#ffffff bg:#333333',
})

#################################################
# Helper 
#################################################
def welcome(): 
    print("==============================")
    print("  Welcome to Scribble Enrich ")
    print("=============================")
    print("\n\n") 

def read_globalconf(): 
    confpath = os.path.expanduser("~/.enrich.json")
    if os.path.exists(confpath): 
        conf = json.loads(open(confpath).read())
    else: 
        conf = {
            'workspaces': [ 'enrich-data'] 
        }
    return conf 

def write_globalconf(conf): 
    confpath = os.path.expanduser("~/.enrich.json")
    with open(confpath, 'w') as fd: 
        fd.write(json.dumps(conf, indent=4))

#################################################
# Validators 
#################################################
class NameValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0: 
            raise ValidationError(message='Customer name should not be empty', 
                                  cursor_position=len(document.text)) # Move cursor to end of input.

class DescValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0: 
            raise ValidationError(message='Customer description should not be empty', 
                                  cursor_position=len(document.text)) # Move cursor to end of input.

class LogoValidator(Validator):
    def validate(self, document):
        url = document.text 
        if len(url) == 0: 
            raise ValidationError(message='Customer logo URL should not be empty', 
                                  cursor_position=len(document.text)) # Move cursor to end of input.
        if not url.startswith('http'): 
            raise ValidationError(message='Customer logo URL should be a valid URL (http://...', 
                                  cursor_position=len(document.text)) # Move cursor to end of input.
        if not (url.lower().endswith('png') or 
                url.lower().endswith('gif') or 
                url.lower().endswith('jpg') or 
                url.lower().endswith('ico')): 
            raise ValidationError(message='Customer logo URL should be a valid URL (http://....png', 
                                  cursor_position=len(document.text)) # Move cursor to end of input.

class ValidDir(Validator):
    def validate(self, document):
        if ((len(document.text) == 0) or 
            (not os.path.exists(document.text)) or 
            (not os.path.isdir(document.text))):            
            raise ValidationError(message='Directory specified doesnt exist or is invalid', 
                                  cursor_position=len(document.text)) 

class ValidFile(Validator):
    def validate(self, document):
        if ((len(document.text) == 0) or 
            (not os.path.exists(document.text)) or 
            (not os.path.isfile(document.text))):            
            raise ValidationError(message='File specified doesnt exist or is invalid', 
                                  cursor_position=len(document.text)) 

##########################
# Management
##########################
def create_workspace(workspace, context): 

    # Create the directories
    try: 
        os.makedirs(workspace) 
    except: 
        pass 
        
    for subdir in ['.', 'conf', 'output', 'data', 'transforms']: 
        try: 
            os.makedirs(os.path.join(workspace, 
                                     context['name'], 
                                     subdir))
        except:
            pass 
        
    # Now write the enrich.json 
    thisdir = os.path.dirname(__file__) 
    template = os.path.join(thisdir, 
                            'templates', 
                            'enrich.json.template')

    updated = render(template, context)
    enrichfile = os.path.join(workspace, 
                              context['name'], 
                              'enrich.json') 
    with open(enrichfile, 'w') as fd: 
        fd.write(updated) 
        
    

def prepare_workspace(): 

    welcome() 

    print("We will ask you a few question\n") 

    # => Directory 
    path_completer = PathCompleter() 
    workspace = prompt('> Workspace directory: ',
                       default='enrich-data',
                       style=style,
                       completer=path_completer)

    print("\nWe need some additional input about the customer:\n") 

    validators = {
        'name': NameValidator(), 
        'description': DescValidator(), 
        'logourl': LogoValidator()
    }

    context = {}
    variables = get_variables(template) 
    for v in variables: 
        context[v] = prompt('> ' + v +": ", 
                            validator=validators.get(v,None))

    # => Update the global conf file...
    conf = read_globalconf() 
    if workspace not in conf['workspaces']: 
        conf['workspaces'].append(workspace) 
    write_globalconf(conf) 

def prepare_transform(): 

    welcome() 

    print("Bootstrapping a transform\n") 
    
    conf = read_globalconf() 
    workspaces = conf['workspaces'] 
    
    # => Get workspace 
    word_completer = WordCompleter(workspaces, ignore_case=True)    
    if os.path.exists('enrich-data'): 
        defaultdir = "enrich-data" 
    else: 
        defaultdir = "."

    workspace = prompt('> Workspace: ', 
                       default=defaultdir, 
                       completer=word_completer,
                       validator=ValidDir()) 

    # => Select the customer 
    customers = os.listdir(workspace) 
    if len(customers) == 0: 
        print("Looks like you havent created the workspace properly. Please run bootstrap workspace") 
        return 

    customer_completer = WordCompleter(customers, ignore_case=True)
    customer = prompt('> Customer: ',
                      default=customers[0], 
                      completer=customer_completer) 

    # => Bootstrap a template 
    templatedir = os.path.join(os.path.dirname(__file__), 
                               'templates') 

    # => Load the manifest 
    print("\nPlease provide transform details\n") 
    defaults = { 
        'version': "1.0",
        'minorversion': "1.0"
    }
    validators = {
        'name': NameValidator(), 
        'description': DescValidator()
    }
    manifest_template = os.path.join(templatedir, 'manifest.json.template') 
    context = {}
    variables = get_variables(manifest_template) 
    for v in variables: 
        context[v] = prompt('> ' + v +": ", 
                            default=defaults.get(v,""), 
                            validator=validators.get(v,None))
    node_manifest = render(manifest_template, context)

    readme_template = os.path.join(templatedir, 'README.md.template') 
    node_readme = render(readme_template, context)    

    # Get the right template
    templates = [t for t in os.listdir(templatedir) if t.endswith('.node.template')]
    templates = [t.replace(".node.template","") for t in templates] 
    template_completer = WordCompleter(templates, ignore_case=True)
    node = prompt('> Type of Transform: ',
                  default=templates[0], 
                  completer=template_completer) 
    
    template_file = os.path.join(templatedir, '{}.node.template'.format(node))
    node_content = render(template_file, context)
    
    # => Now you are ready to write 
    transform_dir = os.path.join(workspace, customer, 'transforms', context['name'])
    try: 
        os.makedirs(transform_dir)
    except:
        pass 
        
    with open(os.path.join(transform_dir, '__init__.py'), 'w') as fd: 
        fd.write(node_content) 

    with open(os.path.join(transform_dir, 'manifest.json'), 'w') as fd: 
        fd.write(node_manifest) 

    with open(os.path.join(transform_dir, 'README.md'), 'w') as fd: 
        fd.write(node_readme) 

    # Bye bye
    print("Bootstrapped transform {} in {}".format(context['name'], transform_dir))

def prepare_config(): 

    welcome() 

    print("Bootstrapping a new configuration\n") 
    
    conf = read_globalconf() 
    workspaces = conf['workspaces'] 
    
    # => Get workspace 
    word_completer = WordCompleter(workspaces, ignore_case=True)    
    if os.path.exists('enrich-data'): 
        defaultdir = "enrich-data" 
    else: 
        defaultdir = "."
    workspace = prompt('> Workspace: ', 
                       default=defaultdir,
                       completer=word_completer,
                       validator=ValidDir()) 

    # => Select the customer 
    usecases = os.listdir(workspace)
    if len(usecases) == 0:
        print("Looks like you havent created the workspace properly. Please run bootstrap workspace")
        return

    usecase_completer = WordCompleter(usecases, ignore_case=True)
    usecase = prompt('> Usecase: ',
                      default=usecases[0],
                      completer=usecase_completer)

    templatedir = os.path.join(os.path.dirname(__file__),
                               'templates')
    validators = {
        'name': NameValidator(),
        'description': DescValidator()
    }

    print("\nPlease provide configuration details\n") 
    config_template = os.path.join(templatedir, 
                                   'config.json.template') 
    context = {}
    variables = get_variables(config_template) 
    for v in variables: 
        context[v] = prompt('> ' + v +": ", 
                            validator=validators.get(v,None))
    config = render(config_template, context)

    name = context['name'] 
    cleaned_name = "".join([x if x.isalnum() else "_" for x in name.lower()])
    configfile = os.path.join(workspace, customer, 'conf',"{}.json".format(cleaned_name))
    with open(configfile, 'w') as fd: 
        fd.write(config)

    # Bye bye
    print("Bootstrapped configuration {} in {}".format(context['name'], configfile))
    
def bootstrap(what): 
    
    if what == 'workspace': 
        prepare_workspace()
    elif what == 'transform': 
        prepare_transform()
    elif what == 'config': 
        prepare_config()
    
    
def bootstrap1(config, transform, transformtype): 
    """
    Initialize a new pipeline...
    """
    thisdir = os.path.dirname(__file__) 
    
    # => First ensure that the destination exists 
    if not os.path.exists(config): 
        os.makedirs(config) 
    elif os.path.isfile(config): 
        raise Exception("Config path cannot be a file") 
        

    # => Not install the transform 
    if transform not in ['none', None]: 
            
        transform_path = os.path.join(config, 
                                      'transforms', 
                                      transform)
        
        try: 
            os.makedirs(transform_path) 
        except:
            pass 

        contentpath = os.path.join(thisdir, 
                                   "{}.template".format(transformtype))
        if not os.path.exists(contentpath): 
            raise Exception("Unsupported transform type template. Use generic") 

        content = open(contentpath).read()         
        initfile = os.path.join(transform_path, '__init__.py')
        with open(initfile, 'w') as fd: 
            fd.write(content) 

        # Insert a manifest as well...
        manifest = os.path.join(transform_path, 'manifest.json')
        with open(manifest, 'w') as fd: 
            fd.write(json.dumps({
                "name": "MyModule", 
                "version": 1.0, 
                "minorversion": 1.0,
                "params": [
                    {
                        "name": "Path", 
                        "required": True, 
                        "description": """\
Path for the module such as 
c:\path\to\file 
/path/to/file
"""                       
                    }
                ]
            }, indent=4))

            
        
############################################################
# Convert 
############################################################
def convert(what, src, dst): 
    """
    Take the old style transform skin and convert into a package each.
    """
    if os.path.exists(dst): 
        raise Exception("Destination exists. Please remove it.") 

    if not os.path.exists(src): 
        raise Exception("Source does not exist. Invalid directory") 
        
    basename = os.path.basename(src) 
    try: 
        enrichjson = os.path.abspath(os.path.join(src, '..', '..', 'enrich.json'))
        usecase = json.load(open(enrichjson))
    except: 
        traceback.print_exc() 
        usecase = {
            'org': {
                'name': 'Unknown', 
                'customer': 'Unknown'
            }
        }

    try: 
        manifestcontent = json.load(open(os.path.join(src, 'manifest.json')))
    except: 
        manifestcontent = {} 
            
    # => 
    defaults = { 
        'what': what, 
        'basename': basename, 
        'name': basename, 
        'version': '0.1', 
        'description': basename + " package",
        'author': 'Scribble Data', 
        'email': 'support@scribbledata.io',
        'url': 'https://github.com/scribbledata/scribble-enrichsdk',
        'customer': usecase['org']['name'],
        'usecase': usecase['org']['name']
    }

    params = {} 
    for k, v in defaults.items(): 
        params[k] = manifestcontent.get(k,v)
        
        
    # Prepare the destination 
    os.makedirs(dst) 

    # Copy the code 
    shutil.copytree(src, os.path.join(dst, basename)) 

    # => Write the manifest file 
    # Write the manifest file...
    manifest = """\
include LICENSE
recursive-include %(basename)s *py *.html *md *template *json
"""
    manifest = manifest % params 
    with open(os.path.join(dst, 'MANIFEST.in'), 'w') as fd: 
        fd.write(manifest) 
            
    # => Write license content
    license = """\
All rights reserved (c) Scribble Data          
"""
    with open(os.path.join(dst, 'LICENSE'), 'w') as fd: 
        fd.write(license) 

    # => Write the setup file 
    setupcontent = """\
import os
from setuptools import setup

setup(name='%(basename)s',
      version='%(version)s',
      description='%(description)s',
      url='%(url)s', 
      author='%(author)s',
      author_email='%(email)s', 
      license='None',
      packages=['%(basename)s'],
      zip_safe=True,
      include_package_data=True,
      entry_points={
          '%(customer)s.%(what)s': ['%(basename)s=%(basename)s'],
      },
)
"""
    setupcontent = setupcontent % params 

    with open(os.path.join(dst, 'setup.py'), 'w') as fd: 
        fd.write(setupcontent) 
        
    print("Created a package in {}".format(dst)) 
        



from django.urls import reverse
from enrichsdk.lib.customer import find_usecase

overview_spec = {
    'name': 'Control Center',
    'description': "Implement and Operate Complex Applications",
    "details": """<p class='text-center'>This dashboard gives an overview of the services enabled on this platform. Click on each tab to see details. The purpose of this platform is to enable strong and continuous model-driven operations.</p>

    <p class='text-center'>Product Owner is <a href=''>Sr. Exec Name</a>, and is currently managed by <a href="https://www.scribbledata.io">Scribble Team</a>.</p>

<p class='text-center'>Please reachout to <a href=''>Data Team</a> or <a href='mailto:support@scribbledata.io'>Scribble Support</a> for technical help</p>""",
    'usecase': find_usecase(__file__),
    'force_list': [],
    'usecases': [
        {
            'name': "ComplexApp",
            'category': 'Finance',
            'owner': 'John Deere',
            'description': 'Complex computation',
            'frequency': 'Daily',
            'services': ['Pipeline', 'App'],
            'datasets': ['many_tables'],
            'status': 'Production'
        },
    ],
    'implement': [
        {
            "category": "Tools",
            'label': "SDK",
            'url': '/docs/sdk/index.html',
            "description": "Developer Interface"
        },
        {
            'category': 'Tools',
            'label': 'Catalog',
            'url': 'APPNAME:catalog:index',
            "description": "Data Schemas and Notes"
        },
    ],
    'operate': [
        {
            'category': 'ComplexAnalysis',
            'label': 'PipelineName',
            'customer': 'UsecaseName',
            'pipeline':  'DailyForecaster',
            'description': 'Daily forecast run'
        },
    ],
    "audit": [
    ],
    "access": [
    ],
    "monitor": [
    ]
}

def get_spec():
    return overview_spec



import os
import json
from datetime import datetime, date, timedelta
from dateutil import parser as dateparser
import logging
import papermill as pm
from enrichsdk import Compute, S3Mixin, CheckpointMixin

logger = logging.getLogger("app")

class NotebookExecutorBase(Compute):
    """
    A built-in transform baseclass to handle standard notebook
    operation and reduce the duplication of code.

    Features of this transform include:

        * Support for custom args and environment
        * Support for automatic capture and surfacing of output and err

    Configuration looks like::

         class MyTestNotebook(NotebookExecutorBase):

             def __init__(self, *args, **kwargs):
                 super().__init__(*args, **kwargs)
                 self.name = "TestNotebook"
                 self.notebook = os.path.join(thisdir, "Test-Notebook.ipynb")

             @classmethod
             def instantiable(cls):
                 return True

             def get_environment(self):
                 return {
                     'SECRET': credentials
                 }
        """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "NotebookExecutorBase"
        self.notebook = None
        self._environ = os.environ.copy()

    @classmethod
    def instantiable(cls):
        return False

    def preload_clean_args(self, args):
        """
        Standard args preprocessor. Make sure that
        an artifacts directory is created for storing the
        configuration file, output notebook and stdout/err.

        """
        args = super().preload_clean_args(args)

        # Insert artifacts if not available..
        if 'artifacts' not in args:
            args["artifacts"] = self.get_file("%(output)s/%(runid)s/artifacts", create_dir=True)
            try:
                os.makedirs(args['artifacts'])
            except:
                pass

        return args

    def get_notebook(self):
        """
        Define notebook that must be executed

        Returns:

            str: Path to the notebook
        """
        if ((not hasattr(self, 'notebook')) or
            (self.notebook is None) or
            (not isinstance(self.notebook, str)) or
            (not os.path.exists(self.notebook))):
            raise Exception("Missing notebook. Missing/invalid path: {}".format(getattr(self, 'notebook', '')))

        notebook = self.notebook
        notebook = os.path.abspath(notebook)
        return notebook

    def get_environment(self):
        """
        Pass any additional parameters...
        """
        return {}

    def run_notebook(self, config, configfile):

        inputnb = self.get_notebook()
        name = os.path.basename(inputnb)
        outputnb = os.path.join(config['artifacts'], 'run-' + name)

        # Pass whatever parameters that the notebook can accept
        params = {
            'config': configfile
        }

        outfile = os.path.join(config['artifacts'], name + '.out')
        errfile = os.path.join(config['artifacts'], name + '.err')

        try:
            pm.execute_notebook(
                inputnb,
                outputnb,
                parameters=params,
                progress_bar=False,
                log_output=True,
                stdout_file=open(outfile,'w'),
                stderr_file=open(errfile, 'w')
            )

            if not os.path.exists(outputnb):
                raise Exception("output notebook not generated")

            logger.debug("Executed notebook: Stdout",
                         extra={
                             'transform': self.name,
                             'data': open(outfile).read()
                         })
            logger.debug("Executed notebook: Stderr",
                         extra={
                             'transform': self.name,
                             'data': open(errfile).read()
                         })

        except:
            logger.exception("Unable to run the notebook",
                             extra={
                                 'transform': self.name,
                             })
            raise

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug("{} - process".format(self.name),
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))


        config = self.args
        configfile = os.path.join(config['artifacts'], 'config.json')
        dump = lambda : json.dumps(config, indent=4, default=str)
        with open(configfile, 'w') as fd:
            fd.write(dump())

        logger.debug("Parameters to script",
                     extra={
                         'transform': self.name,
                         'data': "Config: {}\n---\n".format(configfile) + dump()
                     })

        # Update the environ
        _environ = os.environ.copy()
        try:
            # Update the environ
            update = self.get_environment()
            os.environ.update(update)

            # Now run the notebook
            self.run_notebook(config, configfile)

        finally:
            os.environ.clear()
            os.environ.update(_environ)

        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass




"""

Mixin to implement checks. Light weight flexible version of the

"""
import inspect
import sys
import json
from collections import defaultdict

class ReconMixin(object):

    def check_all_columns_match(self, spec, segment, frames, params):
        """
        Check whether all columns match...
        """

        status ={
            "check": "Match columns",
            "status": True,
            "message": ""
        }

        columns = set()
        for df in frames.values():
            columns.update(list(df.columns))

        for name, df in frames.items():
            if columns != set(df.columns):
                status['message'] += f"[{name}] Mismatch\n"

        return status

    def check_shapes_match(self, spec, segment, frames, params):
        """
        Check whether shapes match...
        """
        status ={
            "check": "Match Shapes",
            "status": True,
            "message": "",
            "details": ""
        }

        shapes = set()
        for df in frames.values():
            shapes.add(df.shape)

        if len(shapes) > 1:
            status['status'] = False
            status['message'] = f"Found {len(shapes)}. Expected 1"
            for name, df in frames.items():
                status['details'] += f"[{name}] {df.shape}\n"

        return status

    def default_checker(self, spec, data, checkspec, summary):

        sources = spec['sources']

        # Collect segments
        segments = set()
        for s in sources:
            segments.update(list(data.get(s,{}).keys()))

        # For each segment..
        for segment in segments:

            frames = {}
            status = {
                'checker': "default_checker",
                'check': 'Data Availability',
                'status': True,
                "message": ""
            }
            for s in sources:
                if segment not in data.get(s, {}):
                    status['status'] = False
                    status['message'] += f"[{s}] Missing\n"
                    continue
                df = data[s][segment]
                if df is None:
                    status['status'] = False
                    status['message'] += f"[{s}] Null\n"
                    continue
                frames[s] = df

            summary['segments'][segment].append(status)
            if not status['status']:
                continue

            def get_func(cls, name):

                func = getattr(cls, name, getattr(cls, "check_" + name, None))
                if func is not None:
                    return func

                # Check the caller module
                callermod = sys.modules[cls.__module__]
                func = getattr(callermod, name, getattr(callermod, "check_" + name, None))
                if func is not None:
                    return func
                if name in globals():
                    handler = globals()[name]
                    return handler

                raise Exception(f"Handler {name} could not be found")

            for c in checkspec['checks']:
                if isinstance(c, str):
                    handler = get_func(self, c)
                    params = {}
                elif isinstance(c, dict):
                    handler = c.get('handler', c.get('check'))
                    handler = get_func(self, handler)
                    params = c.get('params',{})
                else:
                    raise Exception("Unknown check format")

                if not callable(handler):
                    raise Exception("Handler not a valid method")

                status = handler(spec, segment, frames, params)
                if isinstance(status, dict):
                    summary['segments'][segment].append(status)
                elif isinstance(status, list):
                    summary['segments'][segment].extend(status)
                else:
                    raise Exception(f"Invalid response from checker method {handler}")

    def get_check_specs(self):
        raise Exception("To be overridden by parent of this mixin")

    def run_recon(self, data):

        # Get the checking specification...
        specs = self.get_check_specs()

        specsummary = []
        for spec in specs:

            name = spec['name']

            # Get the checker specification.
            checks = spec['checks']
            if isinstance(checks, list) and isinstance(checks[0], str):
                checks = [
                    {
                        "checks": checks,
                    }
                ]

            # Summary per-spec
            summary = {
                'spec': spec,
                'segments': defaultdict(list)
            }

            # Go through each checker specification. There could be
            # potentially multiple strategies to check...
            for checkspec in checks:
                checker = checkspec.get("checker", "default_checker")

                if not callable(checker):
                    checker = getattr(self, checker)

                if not callable(checker):
                    continue

                checker(spec, data, checkspec, summary)

            specsummary.append(summary)

        return specsummary




import os
import sys
import copy
import json
import time
import traceback
import logging
import subprocess
import tempfile
import sqlite3
import jsonschema
from functools import partial

import numpy as np
import pandas as pd
from pandas.io import sql

from enrichsdk import Sink
from enrichsdk.lib.misc import *
from enrichsdk.core.frames import PandasDataFrame

logger = logging.getLogger("app")

class SQLExport(Sink):
    """
    Export dataframes into the SQL database. Args specify what and how
    the export should happen.

    The transform args provides the specification:

          * exports: A list of files that must be exported. Each is a
            dictionary with the following elements:

              * name: Name of this export. Used for internal tracking and notifications.
              * filename: Output filename. Can refer to other global attributes such as `data_root`, `enrich_root_dir` etc
              * type: Type of the export. Only `sqlite` supported for now
              * frames: List of frames of the type `pandas` that should
                exported as part of this file

    Example::

        ....
        "transforms": {
            "enabled": [
               ...
               {
                 "transform": "SQLExport",
                  "args": {
                      "exports": [
                        {
                           "type": "sqlite",
                           "filename": "%(output)s/cars.sqlite",
                           "frames": ["cars", "alpha"]
                        },
                       ...
                      ]
                    },
                   ...
               }
            ...
           }
         }
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "SQLExport"
        self.description = "Export frames into SQLite/Other formats"
        self.roles_supported = ['Export']
        self.roles_current = 'Export'

        self.testdata = {
            'conf': {
                'args': {
		            "exports": [
			            {
			                "name": "customerinfo",
			                "filename": "%(output)s/%(runid)s/customerinfo.sqlite",
			                "type": "sqlite",
			                "frames": ["customerinfo"]
			            }
                    ]
                }
            },
            'data': {
                "customerinfo": {
                    "transform": "MemberLoyaltyMetadata",
                    "filename": "customerinfo.csv",
                    "params": {
                        "sep": ",",
                        "dtype": {
                            'MEMBERSHIP_ID_SUFFIX': 'str'
                        }
                    }
                }
            }
        }

    def preload_clean_args(self, args):
        """
        Enforce the args specification given in the
        example above
        """
        args = super().preload_clean_args(args)

        if len(args) == 0:
            raise Exception("Empty args provided")

        if (('exports' not in args) or
            (not isinstance(args['exports'], list))):
            raise Exception("SQLExport requires a series of exports (a list)")

        for e in args['exports']:

            if ((not isinstance(e, dict)) or
                ('filename' not in e) or
                ('name' not in e) or
                ('type' not in e) or
                ('frames' not in e)):
                raise Exception("Each element of the export should be a dictionary with filename, type, and frames")

            if (('replace' in e) and
                (not isinstance(e['replace'], bool))):
                raise Exception("replace is a boolean variable")

            if  e['type'] != 'sqlite':
                raise Exception("Only sqlite exports are supported in current version")

            e['filename'] = os.path.abspath(self.config.get_file(e['filename']))
            e['relpath'] = os.path.relpath(e['filename'], self.config.data_root)

        return args

    def process(self, state):
        """
        Execute the export specification.
        """
        # Sanity check...
        for e in self.args['exports']:

            # Collect all the frames data
            missing = []
            invalid = []
            frames = {}
            for f in e['frames']:
                detail = state.get_frame(f)
                if detail is None:
                    missing.append(f)
                    continue

                if detail['frametype'] != 'pandas':
                    invalid.append(f)
                    continue

                frames[f] = detail

            if len(missing) > 0 or len(invalid) > 0:
                logger.error("Unable to export frames",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': "Invalid: {}\nMissing: {}".format(invalid, missing)
                             }))
                raise Exception("Error while exporting")

            filename = e['filename']
            filename = os.path.abspath(self.config.get_file(filename))
            relpath = self.config.get_relative_path(filename,
                                                    what='enrich_data_dir')

            name = e.get('name', os.path.basename(filename))
            replace = e.get('replace', False)

            try:
                os.makedirs(os.path.dirname(filename))
            except:
                pass

            # Creating a database file
            conn = sqlite3.connect(filename)
            c = conn.cursor()

            for f in frames:

                # => Write the frames
                overrides = self.frame_get_overrides(frames[f])
                override_present = len(overrides) > 0
                save = overrides.get('save', True)
                if save:
                    try:

                        # Status flag
                        failed = False

                        df = frames[f]['df']

                        # Drop existing table if necessary...
                        if replace:
                              c.execute(f'DROP TABLE IF EXISTS {f}')

                        # => First create the table schema
                        ddl = pd.io.sql.get_schema(df, f)
                        c.execute(ddl) # CREATE table

                        # => Dump the dataframe to a csv
                        fd, tmpfile = tempfile.mkstemp(prefix="sqlexport")
                        df.to_csv(tmpfile, header=False, index=False)

                        # => Load it into sqlite
                        cmd = ["/usr/bin/sqlite3", filename,
                               '-cmd', '.separator ,',
                               ".import {} {}".format(tmpfile, f)]

                        process = subprocess.Popen(cmd,
                                                   stdout=subprocess.PIPE,
                                                   stderr=subprocess.PIPE)
                        out, err = process.communicate()
                        err = err.decode("utf-8")
                        out = out.decode("utf-8")

                        # Dump what we have seen
                        if len(err) > 0:
                            failed = True
                            logfunc = partial(logger.error, "Unable to export {}".format(f))
                        else:
                            logfunc = partial(logger.debug, "Exported {}".format(f))
                        logfunc(extra=self.config.get_extra({
                            'transform': self.name,
                            'data': "Filename:{}\nOutput\n-----\n{}\n\nErr\n----\n{}".format(filename, out, err)
                        }))

                        # => Update the state for this transform..
                        if not failed:
                            state.update_frame(f, {
                                'df': frames[f]['df'],
                                'frametype': frames[f]['frametype'],
                                'transform': self.name,
                                'history': [{
                                    'log': 'Exported to SQLite'
                                }],
                                'params': {
                                    'type': 'lineage',
                                    'transform': self.name,
                                    'dependencies': [
                                        {
                                            'type': 'dataframe',
                                            'nature': 'input',
                                            'objects': [
                                                f
                                            ]
                                        },
                                        {
                                            'type': 'file',
                                            'nature': 'output',
                                            'objects': [
                                                filename
                                            ]
                                        }
                                    ]
                                }
                            })

                    except:
                        logger.exception("Unable to export dataframe {}".format(f),
                                         extra=self.config.get_extra({
                                             'transform': self.name,
                                         }))

                    # Cleanup...
                    try:
                        if os.path.exists(tmpfile):
                            os.remove(tmpfile)
                    except:
                        pass

                    # Dont proceed
                    if failed:
                        raise Exception("Error while exporting {}".format(f))

                else:
                    logger.warn("Did not save {} due to overrides".format(f),
                                 extra=self.config.get_extra({
                                     'transform': self.name,
                                     'data': "Overrides: {}".format(overrides)
                                 }))

            conn.close()
            if not os.path.exists(filename):
                logger.error("SQLite file not created or missing",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': "Filename: {}\nOverride Present: {}".format(filename, override_present)
                             }))
                raise Exception("SQLite file missing")

            # Now create the state entry
            detail = {
                'df': None,
                'frametype': 'db',
                'transform': self.name,
                'params': {
                    'filename': filename,
                    "action": "output",
                    "frametype": "db",
                    "descriptions": [
                        "SQLite export of {} frames ({})".format(len(frames), ",".join(frames))
                    ],
                    "notes": [
                        "Frames included: {}".format(','.join(frames))

                    ],
                    'components': [
                        {
                            'filename': relpath,
                            'type': 'sqlite',
                            'sha256sum': get_checksum(filename),
                            'filesize': "{0:0.3f} MB".format(get_file_size(filename)/(1024*1024)),
                            "modified_time": str(time.ctime(os.path.getmtime(filename))),
                            "create_time": str(time.ctime(os.path.getctime(filename))),

                        }
                    ],
                },
                'history': [
                    {
                        'transform': self.name,
                        'log': "Write SQLite export"
                    }
                ]
            }
            try:
                state.update_frame(name, detail, create=True)
                state.make_note("Generated database export")
            except:
                logger.exception("Unable to store state",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Error while storing")

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass

provider = SQLExport



import re
import os
import sys
import py_compile
import ast 
from setuptools import setup, find_packages 

_version_re = re.compile(r'VERSION\s+=\s+(.*)')

thisdir = os.path.dirname(__file__)
readme = open(os.path.join(thisdir, 'README.rst')).read()

with open('enrichsdk/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(f.read().decode('utf-8')).group(1)))
    
setup(name='enrichsdk',
      version='3.6.2',
      description='Enrich Developer Kit',
      long_description=readme, 
      url='http://github.com/pingali/scribble-enrichsdk',
      author='Venkata Pingali',
      author_email='pingali@scribbledata.io',
      license='All rights reserved',
      scripts=[
      ],
      packages=find_packages(),
      include_package_data=True, 
      zip_safe=False,
      install_requires=[
          'docutils<0.16',
          'click==7.1.2',
          'aiobotocore==1.2.1',
          'sphinx-click>=2.3.0',
          'glob2==0.7',
          'requests>=2.21.0',
          'requests-oauthlib==0.8.0',
          'pytest>=4.6.0',
          'numpydoc>=0.7.0',
          'pandas==0.25.3',
          'dask==2.30.0',
          'distributed==2.30.1',
          'idna==2.8',
          'coverage==5.5',
          'flake8==3.9.0',
          'raven==6.6.0',
          'python-json-logger==0.1.8',
          'python-dateutil==2.8.1',
          's3fs==0.5.1',
          'fsspec==0.8.5',
          'colored==1.3.5',
          'flask-multistatic==1.0',
          'humanize==0.5.1',
          'pytz==2020.1',
          'Flask==1.1.2',
          'Jinja2>=2.10.1',
          'pytest-cov',
          'Markdown>=2.9.10',
          'prompt-toolkit>=2.0.10',
          'pyarrow>=0.9.0',
          'cytoolz>=0.9.0.1',
          'jsonschema>=3.2.0',
          'flask_cors',
          #'moto>=1.3.14',
          'prefect==0.15.7',
          "distro>=1.4.0",
          "gcsfs==0.6.0",
          "jupyter-core>=4.6.1",
          "nbformat>=5.1.2",
          'tzlocal>=2.0.0',
          'texttable',
          'pykafka',
          'redis',
          'gitpython',
          'logstash_formatter',
          'pyhive',
          'pyfiglet',
          'sqlalchemy>=1.4.0',
          'kafka-python==2.0.2',
          'pykafka==2.8.0',
          'papermill>=2.2.2',
          'sqllineage'
      ],
      entry_points = {
          'console_scripts': [
              'enrichpkg=enrichsdk.scripts.enrichpkg:main',
          ],
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "Operating System :: OS Independent",
      ],
      python_requires='>=3.6',
)



"""
Dataframe wrapper. Standardizes interfaces across multiple
implementations of dataframes including pandas, Spark RDD, 
and even dictionaries..
"""
import os, sys, json, glob, copy, time 
import numpy as np 
import pandas as pd 

from datetime import datetime 

__all__ = ['DataFrame', 'PandasDataFrame', 'DictDataFrame']

class DataFrame(object): 
    
    @staticmethod
    def supported(filename):
        raise Exception("Unsupported") 

    @staticmethod
    def read(spec, state): 
        """
        Load the tables...
        """
        return state

    @staticmethod
    def duplicate(df): 
        raise Exception("Unsupported")        
        
class PandasDataFrame(DataFrame): 

    @staticmethod 
    def rename(df, columns): 
        return df.rename(columns=columns) 

    @staticmethod 
    def unstack(df, level, **kwargs): 
        return df.unstack(level, **kwargs) 

    @staticmethod 
    def sort_values(df, columns, **kwargs): 
        return df.sort_values(columns, **kwargs) 

    @staticmethod 
    def to_datetime(df, column, **kwargs): 
        df[column] = pd.to_datetime(df[column], **kwargs) 
        return df 

    @staticmethod 
    def apply(df, func, **kwargs): 
        return df.apply(func, **kwargs) 

    @staticmethod 
    def drop(df, columns): 
        return df.drop(columns, axis=1, inplace=False) 
        
    @staticmethod 
    def reset_index(df, **kwargs): 
        return df.reset_index(**kwargs) 

    @staticmethod 
    def groupby(df, columns): 
        return df.groupby(columns) 

    @staticmethod 
    def apply_column(df, column, func, inplace=True, **kwargs): 
        if inplace: 
            df[column] = df[column].apply(func, **kwargs) 
            return df 
        else: 
            return df[column].apply(func, **kwargs) 

    @staticmethod 
    def apply_row(df, func): 
        df = df.apply(func, axis=1)
        return df 

    @staticmethod 
    def mean(df): 
        return df.mean() 

    @staticmethod 
    def std(df): 
        return df.std() 
        
    @staticmethod 
    def dtypes(df): 
        return df.dtypes 
        
    @staticmethod 
    def shape(df): 
        return df.shape 

    @staticmethod 
    def columns(df): 
        return df.columns

    @staticmethod 
    def head(df,num=5): 
        return df.head(num)

    @staticmethod 
    def slice(df,columns): 
        return df[columns] 

    @staticmethod 
    def fillna(df, column, value): 
        df[column] = df[column].fillna(value) 
        return df 

    @staticmethod 
    def astype(df, column, value): 
        df[column] = df[column].astype(value) 
        return df 

    @staticmethod 
    def merge(left, right, **kwargs):
        return pd.merge(left, right, **kwargs) 
    
    @staticmethod 
    def nlargest(df, n, columns, **kwargs): 
        return df.nlargest(n, columns,**kwargs) 

    @staticmethod
    def slice_columns(df, columns): 
        return df[columns] 

    @staticmethod
    def tolist(df, transpose=False): 
        return df.values.tolist()


    @staticmethod
    def rank(df, column, **kwargs): 
        return df[column].rank(**kwargs) 

    @staticmethod 
    def get_schema(df, table_name): 
        return pd.io.sql.get_schema(df, table_name) 

    @staticmethod 
    def get_generic_dtype(df, c): 
        d = df.dtypes[c] 
        d = str(d) 
        if ((d not in ['int64', 'float64', 'int8', 'bool', 'category']) and 
            (not d.startswith('datetime'))): 
            d = 'str'
        return d 

class DictDataFrame(DataFrame): 

    pass 




import os
import sys
import json
import re
import copy
from enrichsdk.contrib.lib.transforms.fileops import FileOperationsBase
import logging

logger = logging.getLogger("app")

__all__ = ['FileOperations']

class FileOperations(FileOperationsBase):
    """
    FileOperations performs a number of operations on files
    generated by pipelines.

    The transform takes a list of actions. The only action type
    supported for now is `copy`. Each copy task requires source,
    destination, and instruction on what to do with existing file.

    Example::

        {
                "transform": "FileOperations",
                "enable": true,
                "dependencies": {
                   ....
                },
                "args": {
                    "actions": [
                          {
                            "action": "copy",
                            "src": "%(output)s/%(runid)s/profile.sqlite",
                            "dst": "%(data_root)s/shared/campaigns/profile_daily/profile.sqlite",
                            "backupsuffix": ".backup"
                          }
                     ]
                }
        }

    """
    @classmethod
    def instantiable(cls):
        """
        Make this class instantiable
        """
        return True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "FileOperations"
        self.description = "File operations such as move/copy etc."

        test_root = os.environ['ENRICH_TEST']
        self.testdata = {
            'data_root': os.path.join(test_root, self.name),
            'outputdir': os.path.join(test_root, self.name, 'output'),
            'inputdir': test_root,
            'statedir': os.path.join(test_root, self.name, 'state'),
            'global': {
                'args': {
                    'rundate': '2020-01-10'
                }
            },
            'conf': {
                'args': {
                    "actions": [
                        {
                            "action": "copy",
                            "src": "%(output)s/%(runid)s/outputs/cars.csv",
                            "dst": "%(data_root)s/shared/%(rundate)s/hello.csv"
                        }
                    ]
                }
            },
            'data': {
            }
        }


provider = FileOperations
